// OK1DXD Precise Automatic Noise Figure Meter
// Options installed:
//    Internal 10bit AD converter with LM336-2.5V EXTERNAL reference
//    AD 8306 LOG detector: 2.5V max output
//    Temperature sensor: Dallas ds18b20
//    I2C 16*2 LCD display
//    Noise head On/Off control
//    Internal EEPROM used to store calibration values Y, P_delta, Temperature
//    LOG detector calibration table
// --------TO DO ---------------------------
// 2nd Stage contribution calculation
// Store NoiseHead ENR by band
// service mode onboard jumper
// add more points on calibration table
// (c) OK1DXD 2017, 2018

float ver = 5.0;  // SW version

//EXTERNAL ADC reference - put a measured value on ref pin here:
float refVolts = 2.494;  // LM336-2.5V External reference voltage calibration[Volts]
int servicemode = 1;     // go to the service when flag = 1 (on board jumper?)
float powerdB   = 0.00;  // power in dBm

#include <OneWire.h>
#include <DallasTemperature.h> // temperature sensor
#define ONE_WIRE_BUS 3
#define TEMPERATURE_PRECISION 11
//
#include <EEPROM.h> // internal EEPROM emulation used
#include <math.h>
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
LiquidCrystal_I2C lcd(0x27,16,2); // I2C display ID = 0x27

float temp = 17.0;   // current room temperature
float c_temp = 20.5; // Temperature just when did calibration
float t_delta = 4.0; // treshold [C] for recalibration need (when current temperature differs from stored one)
float voltage = 0.0;
float enr = 7.08;    // ENR for 145MHz
//float enr = 8.01;     // ENR for 145MHz set here
long  coldV = 0UL;    // initial voltage level when noise off (cold)
long  hotV  = 0UL;    // initial voltage level when noise on (hot)
double dBcold = -20;  // initial power level when noise off (cold)
double dBhot  = -26;  // initial power level when noise on  (hot)
double P_off = 0;     // noise power/ head off {W}
double P_on =  0;     // noise power/ head on  {W}
float C_delta = 0.0000014327 ; // = (P_on - P_off) noise head calibration initial value
float P_delta = 0.0 ; // = (P_on - P_off) DUT value
float gain = 0;
int maxcycles = 70;   // number of measuring cycles (multipl. by 100 readings)
int j = 0;            // cycle counter

// Y, NF equation used:
float y = pow(10, abs((dBhot - dBcold) / 10)); // Y factor
float nf = enr - 10 * (log10(y - 1));          // Noise figure
float dBm = 0.0;
float tr = (temp + 273.15);   // Room temperature [K] - input from sensor
float te = tr * (1 + pow(10, enr / 10)); // Excesive temperature (ENR7:1770K/145MHz)
float th = (te - y*tr) / (y - 1); //Thot noise temperature calculation

// variables for system calibration
float ys = 5.50 ;      // system Ys from calibration
float ts = (te-ys*tr)/(ys-1);  // initial Ts
int eeAddress = 0;      // EEPROM adress for Ys calibration value storing
long adc = 0UL;            // reading from the ADC, an unsigned loong integer
bool head = false;      // noise source on-off toggle
int pinHead = 2;        // noise Head on-off digital pin
int pinADC = 2 ;        // ADC analogue reading pin
int pinCAL = 3 ;        // CALIBRATION/Measuring switch pin
int mode   = 4 ;        // Monitor mode onboard jumper

//calibration tables dBm, miliVolts
int frq[] = {28, 145, 432}; //frequency bands used [MHz]
int frq_index = 1 ;  // set cursor on the 145MHz as default FRQ
int inmV = 1000;     // input value for testing purpose only
int maxnumber = 12;  // number of calibration points
int db[] = { -90, -80, -70, -60, -50, -45, -40, -30, -20, -10, 0, 10};

// LOG detector calibration table
int mv[3][12] = {
 //-90, -80, -70, -60,  -50,  -45,  -40,  -30,  -20,  -10,  0,   10};
  {412, 453, 615, 814, 1009, 1109, 1206, 1405, 1609, 1818, 2025, 2224}, //   28MHz
  {412, 453, 615, 814, 1009, 1109, 1206, 1405, 1609, 1818, 2025, 2224}, //   145MHz
  {412, 453, 615, 814, 1009, 1109, 1206, 1405, 1609, 1818, 2025, 2224}, //   432MHz
  };
int i = 0;
// Temperature measurement setup
//int temp_pin = 3; // pin for DS18b20 temperature sensors, must be pulled up by external resistor
OneWire oneWire(ONE_WIRE_BUS);
// declare as sensor reference by passing oneWire reference to Dallas Temperature.
DallasTemperature sensors(&oneWire);
// SENSOR ADDRESS ID:
DeviceAddress tempSensor01 = {0x28, 0xFF, 0xF8, 0x0C, 0xB4, 0x16, 0x05, 0x99};
//******************************************************************************************
void setup(void)
{
  pinMode(pinCAL, INPUT);     //ADC input CAL. switch
  pinMode(pinADC, INPUT);     //ADC input from LOG detector
  pinMode(pinHead, OUTPUT);   //digital output
  digitalWrite(pinHead, head);  //Set the noise head OFF (false) at the start
 
  // Very important note – when using an external voltage reference, you must set the analogue
  // reference to EXTERNAL before using ANY (!) analogRead
  analogReference(EXTERNAL); // use AREF for reference voltage
  adc = analogRead(pinADC);  // one Analog reading for sure needed!

  Serial.begin(9600);  // Serial Monitor output for servicemode output
  sensors.begin();  // Dallas TMP sensor(s)
  sensors.setResolution(tempSensor01, TEMPERATURE_PRECISION);
  
  /* un-rem this for the first time EEPROM initialization only 
  EEPROM.put(eeAddress, 5.1); // Ys initial EEPROM value test only
 // eeAddress = sizeof(float); //Move address to the next byte after float Ys
  EEPROM.put(sizeof(float), 0.0000014327); //P delta initial EEPROM value test only
 // eeAddress =  eeAddress + sizeof(float); // move pointer to the next float memory cell
  EEPROM.put((2*sizeof(float)), 20.0); //temperature initial EEPROM value only
  */
  // Reads stored system data from EEPROM at system start
  EEPROM.get(eeAddress, ys);     // Reads stored system Ys data
  EEPROM.get(sizeof(float), C_delta ); // Read stored system Power delta
  EEPROM.get((2*sizeof(float)), c_temp ); // Read last calibration outside temperature
  
  // Welcome screen
  lcd.init();
  lcd.backlight();
  lcd.begin(16, 2);
  lcd.print("NF Meter V:");
  lcd.print(ver, 1);
  lcd.setCursor(0, 1);
  lcd.print("Ys:");
  lcd.print(ys, 1); 
  lcd.print(" Ts:");
  ts = (te-ys*tr)/(ys-1);  // calculate Ts
  lcd.print(ts, 0);
  lcd.print("K");
  delay(2000);
  lcd.clear();
}
//******************************************************************************************
// Main loop
void loop(void)
{
  while (analogRead(pinCAL)> 50 ) // if Calibration switch open
      { // Go to the System Calibration mode
      lcd.setCursor(0, 0);
      lcd.print("Calibrat.       ");
      // get & display head temperature here later
      measure();
      lcd.setCursor(0, 1);
      lcd.print("                ");
      lcd.setCursor(0, 1);
      lcd.print("Ys:");
      lcd.print(y, 1);
      ts = (te-y*tr)/(y-1);  // calculate Ts
      lcd.setCursor(7, 1);
      lcd.print(" Ts:");
      lcd.print(ts, 0);
      lcd.print("K");
      if (servicemode == 1) serialmonitor(); // call service mode monitor if flag
      EEPROM.get(eeAddress,ys); // last stored system Y
      //
      if (y != ys) {    // store only if Y was changed to protect EEPROM life
          EEPROM.put(eeAddress,y);
          eeAddress = sizeof(float); //Move address to the next byte after float
          // P_ delta calculation
          EEPROM.put(sizeof(float),P_delta);// Store new system gain also
          EEPROM.put((2*sizeof(float)),temp); // stored outside temperature when do calibration
          lcd.setCursor(0, 0);
          lcd.print("Storing Ys & Temp"); 
          delay(2000);
          } 
      }   // end of while calibration cycle 
  lcd.setCursor(0, 0);
  lcd.print("                "); // clear first line only
  lcd.setCursor(0, 0);
  lcd.print("Measuring");
  measure ();
  if (servicemode == 1) serialmonitor(); // call service mode monitor
  
  // was Temperature changed from the last calibration?
  EEPROM.get((2*sizeof(float)),c_temp ); // get stored outside temperature when did last calibration
  if (abs(temp - c_temp) > t_delta) {   // compare to current temperature (a treshold considered)
    lcd.clear();
    lcd.print("TemperatureIsOut");
    lcd.setCursor(0, 1);
    lcd.print("!DO CALIBRATION!");
    delay(3000);
    lcd.clear();
    } 
  show_nf();
  delay(2000);
}
//******************************************************************************************
// Functions Declaration
//******************************************************************************************
//
// get average voltage [mV] from input pinADC,
float getADCreading() {
  int maxloop = 100;    //nr of loop cycles for averaging
  int i = 0;            //index
  adc = 0UL;            // clear
  voltage = 0;
  for (i = 0; i < maxloop; i++)  {
    adc = (adc + analogRead(pinADC));     //one Analog reading
   delayMicroseconds(50);
    }
  adc = adc / maxloop;  // get an AVERAGE value
  voltage = (adc * refVolts) / 1023; // Convert the ADC result to Volts in Vout
  inmV = (voltage * 1000); // Convert to integer milivolts
  dBm = getdBValue(inmV); // Get dB value from calibration table
  return dBm;
}

// getdBValue returns value from the CALIBRATION table on given frq_index
float getdBValue(int inputmV) {
  int xA;
  int xB;
  int dx; //delta
  int dy;
  uint8_t index;
  float k; //slope
  for (int i = 0; i < maxnumber; i++) // number of calibration points
  {
    if (mv[frq_index][0] >= inputmV) // input value is under the detector range
    {
      return db[0];
      break; // done
    }
    if (mv[frq_index][i] > inputmV) // find the higher point = index
    {
      xA = mv[frq_index][i - 1];
      xB = mv[frq_index][i];
      index = i;
      break; // done
    }
  }
  dx = xB - xA;
  dy = db[index] - db[index - 1];
  k = (float) dy / (float) dx;
  return k * (float) (inmV - xA) + db[index - 1];
}

// Set On/Off Noise head
boolean noiseHead(boolean head)  {
  digitalWrite(pinHead, head);  //Set the noise source: head=false -> OFF , head=true -> ON
  delay(10);                    // wait a little bit to seattle down
}

// one measurement on-off cycle
void measure () {
  dBcold = 0; // clear value
  dBhot  = 0;
  for (j = 0; j < maxcycles; j++) {
    noiseHead(head = true); // noise head ON
    dBhot = dBhot + getADCreading();
   // 
    noiseHead(head = false); // noise head OFF
    dBcold = dBcold + getADCreading();
    lcd.setCursor((90 + j) / 10, 0); // move & print dot progress bar one position right
    lcd.print(".");
  }
  
  dBcold = dBcold / maxcycles;
  P_off = pow (10, (dBcold / 10));
  dBhot = dBhot / maxcycles ;
  P_on  = pow (10, (dBhot / 10));
  P_delta = (P_on - P_off); //P_delta from DUT
  gain = 10* log10((P_delta)/ C_delta); 
  y = pow(10, abs((dBhot - dBcold) / 10));  // Y calculation
 
// Y, NF Rounded to 2 decimal places:
  float bar = y * 10.0; // bar now = 12
  y  = round (bar );    // y now = 1.2 or something very close but that's how FP is and always has been.
  y = y/10 ;
  nf = enr - 10 * (log10(y - 1)); // NF DUT
  
  float bar2 = nf * 10.0;
  nf = round (bar2 );    // need to check as roundig is doing well...
  nf = nf/10 ;
  //
  sensors.requestTemperatures();
  temp = sensors.getTempC(tempSensor01); // get temperature [C]
  tr = (temp + 273.15);  // current room temperature contribution
  te = tr * (1 + pow(10, enr / 10)); // Excesive temperature (ENR7:1770K/145MHz at 290K)
  th = (te - y * tr) / (y - 1);      // T DUT (ON) noise temperature
}
// Display a values on LCD
void show_nf() {
  lcd.setCursor(0, 0);
  lcd.print("G = ");
  lcd.setCursor(3, 0);
  lcd.print("     ");
  lcd.setCursor(3, 0);
  lcd.print(gain, 1);
  //
  lcd.setCursor(0, 1);
  lcd.print("NF= ");
  lcd.setCursor(3, 1);
  lcd.print("      ");
  lcd.setCursor(3, 1);
  lcd.print(nf, 1);
}
// serial monitor on PC screen
void serialmonitor ()
{
  if (head) Serial.print("OFF: ");
  else Serial.print("ON/ ");
  Serial.print("TMP=");
  Serial.print(c_temp, 1); // temperature at last calibration
   Serial.print("/");
  Serial.print(temp, 1);   // current measured temperature
  Serial.print(" Ys=");    //System Ys (from calibration table)
  Serial.print(ys, 2);
//  Serial.print(" Ts=");  //System noise temperature (calculated from Ys)
//  Serial.print(ts, 0);
  Serial.print(" Delta=");  // System power delta (from calibration table) for gain calculation
  Serial.print(C_delta, 8);
  Serial.print(" K dBcold=");
  Serial.print(dBcold, 1);
  Serial.print(" dBhot");
  Serial.print(dBhot, 1);
  Serial.print(" Gain="); // DUT gain
  Serial.print(gain, 1);
  Serial.print(" Y="); // DUT Y
  Serial.print(y, 2);
  Serial.print(" NF="); // DUT NF
  Serial.print(nf, 2);

  Serial.print(" Te="); // Noise Head Temperature (ON) calculated from ENR
  Serial.print(te, 0);
  Serial.print("K T_DUT="); // DUT noise temperature (ON)
  Serial.print(th, 0);
  // ADC reading for noise on
  Serial.print("K pinADC="); // ADC RAW
  Serial.print(adc);
  Serial.print("/U=");
  Serial.print(voltage * 1000, 0); // ADC mV input
  Serial.print("mV");
  Serial.print(" pinCAL=");
  Serial.print(analogRead(pinCAL));  // Calibration/Measuring switch reading
  Serial.print("(RAW)");
  Serial.println("");

  // LOG det readings on LCD OUTPUT also here:
  lcd.setCursor(8, 0);
  lcd.print(" C:     ");
  lcd.setCursor(11, 0);
  lcd.print(dBcold, 1);

  lcd.setCursor(9, 1);
  lcd.print("H:     ");
  lcd.setCursor(11, 1);
  lcd.print(dBhot, 1);
}
